package Channel;
use strict;

# Channel.pm
#
# Class for channel-related operations
#

our $log2 = log(2);

# Constructor ($matrix)
#   $matrix: the channel matrix. It should be a ref to an array of the rows, each of which is a ref to an
#             array of values. eg  [ [1,2,3], [4,5,6] ]
#
sub new {
	my ($class, $matrix) = @_;
	my $self = { matrix => $matrix };
	return bless $self;
}

sub new_symmetric {
	my ($class, $size, $bigval) = @_;
	my $lowval = (1 - $bigval) / ($size-1);

	my $c = [];
	for my $i (0..$size-1) {
		push @$c, [];
		for my $j (0..$size-1) {
			$c->[$i][$j] = $i == $j ? $bigval : $lowval;
		}
	}

	my $self = { matrix => $c };
	return bless $self;
}

# input_no ()
#   Returns the number of input values of the channel (the number of rows of the matrix)
#
sub input_no {
	my ($self) = @_;
	return scalar @{$self->{matrix}};
}

# output_no ()
#   Returns the number of output values of the channel (the number of columns of the matrix)
#
sub output_no {
	my ($self) = @_;
	return scalar @{$self->{matrix}[0]};
}

# test_matrix ()
#   Performs tests to check that the matrix is valid (for the moment it checks that the rows add up to one)
#   Returns 1 if the tests succeed, undef otherwise
#
sub test_matrix {
	my ($self) = @_;
	my $m = $self->{matrix};

	for my $i (0..$self->input_no-1) {
		for my $j (0..$self->output_no-1) {
			defined $m->[$i][$j]
				or return undef, "cell $i, $j is undefined";
		}
	}

	for my $i (0..$self->input_no-1) {
		my $row = $m->[$i];
		my $s = sum($row);
		$s > .9999 && $s < 1.0001
			or return undef, "sum of row $i is $s";
	}
	return 1;
}

# capacity ($e)
#   Computes the capacity of the channel using the Arimoto-Blahut algorithm
#   $e is the accuracy to which the capacity should be computed
#
sub capacity {
	my ($self, $e) = @_;
	$e ||= 0.001;		# default value
	my $matrix = $self->{matrix};

	my $m = scalar @{ $matrix };		# no of X's values
	my $n = scalar @{ $matrix->[0] };	# no of Y's values

	my @F = map 0, 1..$m;
	my @Px = map 1/$m, 1..$m;
	my @Py = map 0, 1..$n;

	while(1) {
		# update Py
		for my $k (0..$n-1) {
			$Py[$k] = 0;
			$Py[$k] += $matrix->[$_][$k] * $Px[$_]
				for 0..$m-1;
		}

		# update F
		for my $j (0..$m-1) {
			my $s = 0;
			# NOTE: this is e-base log, not 2-base!
			$s += $matrix->[$j][$_] > 0
				? $matrix->[$j][$_] * log( $matrix->[$j][$_] / $Py[$_])
				: 0
				for 0..$n-1;
			$F[$j] = exp($s);
		}

		# check stop condition
		my $x = 0;
		$x += $F[$_] * $Px[$_]
			for 0..$m-1;
		my $IL = log2($x);
		my $IU = log2(max(@F));

		return ($IL, \@Px)		if $IU - $IL < $e;

		# update Px
		$Px[$_] *= $F[$_] / $x
			for 0..$m-1;
	}
}

# capacity_sym ()
#   Computes the capacity of the channel assuming it is symmetric (in this case C = log|O| - H(row)
#   NOTE: symmetry is assumed, we don't test it
#
sub capacity_sym {
	my ($self) = (@_);
	my $row = $self->{matrix}[0];

	my $s = 0;
	$s += $_ == 0 ? 0 : - $_ * log2($_)
		for @$row;

	return log2(scalar @$row) - $s;
}

# capacity_partsym ($indexes)
#   Computes the capacity of the channel assuming that it is partially symmetric (some columns
#   are constant, the rest are symmetric). It uses the formula from the channels paper.
#   $indexes is an array with the number of the columns
#   that form the symmetric part. No testing of symmetry is performed
#
sub capacity_partsym {
	my ($self, $indexes) = @_;
	my $row = $self->{matrix}[0];

	# calculate the sum and the entropy of the symmetric part of a row
	my ($entr, $sum) = (0, 0);
	for(@$indexes) {
		my $val = $row->[$_];
		$sum += $val;
		$entr += $val == 0 ? 0 : - $val * log2($val);
	}

	return $sum * log2(@$indexes / $sum) - $entr;
}

sub probable_innocence {
	my ($self) = @_;
	my $matrix = $self->{matrix};

	my $n = scalar @{ $matrix };		# no of rows
	my $m = scalar @{ $matrix->[0] };	# no of cols

	for my $z (0..$m-1) {
		for my $i (0..$n-1) {
			my $test = ($n-1) * $matrix->[$i][$z];
			for my $j (0..$n-1) {
				# check that (n-1)p(o_z | a_i) >= p(o_z | a_j)
				return undef, $z, $i, $j
					if $test < $matrix->[$j][$z];
			}
		}
	}
	return 1;
}

# prob_error ($x)
#
# Returns the bayesian probability of error, using $x as the input distribution
#
sub prob_error {
	my ($self, $x) = @_;

	my $m = $self->{matrix};
	my $s = 0;
	for my $i(0.. @{$m->[0]}-1) {
		my $max = 0;
		for my $j(0..@$m-1) {
			$max = $m->[$j][$i] * $x->[$j]
				if $max < $m->[$j][$i] * $x->[$j];
		}
		$s += $max;
	}
	return 1-$s;
}

sub santhi_vardy {
	my ($self, $x) = @_;

	my $c = $self->{matrix};
	my $n = @$c;
	my $m = @{$c->[0]};

	my $s = 0;
	for my $j (0..$m-1) {
		#my $xo = post($c, $x, $j);
		my $xo = [ map $x->[$_] * $c->[$_][$j], 0..@$c-1 ];		# $x .* the j-th column
		my $sum = 0; $sum += $_ for @$xo;
		if($sum > 0) {
			$_ /= $sum for @$xo;
		}

		for my $i (0..$n-1) {
			next unless $xo->[$i];
			$s -= $x->[$i] * $c->[$i][$j] * log2($xo->[$i]);
		}
	}

	return 1-2 ** -$s;		# 1 - 2^(-H(A|O)) bound
	#return $s / 2;			# 1/2 H(A|O) bound
}

# to_string
#    Returns a string representation of the channel (useful for printing)
#
sub to_string {
	my ($self) = @_;

	my $s = "";
	for my $row (@{ $self->{matrix} }) {
		$s .= join('  ', map {defined $_ ? sprintf("%.5f", $_) : "undef  "} @$row) . "\n";
	}

	return $s;
}


# Auxiliary functions
#
sub log2 {
	return log($_[0]) / $log2;
}

sub max {
	my $m = $_[0];
	$_ > $m && ($m = $_)
		for @_;
	return $m;
}

sub sum {
	my $row = shift;
	my $s = 0;
	$s += $_	for @$row;
	return $s;
}

sub entropy {
	my ($row) = (@_);

	my $s = 0;
	$s += $_ == 0 ? 0 : - $_ * log2($_)
		for @$row;

	return $s;
}


1;
